from mock import Mock
from cg.apps.tb.dto.summary_response import AnalysisSummary
from cg.services.orders.order_status_service.order_status_service import OrderStatusService
from cg.services.orders.order_status_service.dto.order_status_summary import OrderSummary
from cg.store.models import Order


def test_get_status_summaries(
    order_status_service: OrderStatusService, order: Order, analysis_summary: AnalysisSummary
):

    # GIVEN a store with orders

    # GIVEN that the analysis client returns analysis summaries for the order
    order_status_service.analysis_client = Mock()
    order_status_service.analysis_client.get_summaries.return_value = [analysis_summary]

    # WHEN getting status summaries
    summaries: list[OrderSummary] = order_status_service.get_status_summaries([order.id])

    # THEN the summaries should be returned
    assert summaries



from pathlib import Path
import re
import shutil
import uuid

from cg.services.fastq_file_service.exceptions import ConcatenationError


def concatenate_forward_reads(directory: Path) -> Path | None:
    fastqs: list[Path] = get_forward_read_fastqs(directory)
    if not fastqs:
        return
    output_file: Path = get_new_unique_file(directory)
    concatenate(input_files=fastqs, output_file=output_file)
    validate_concatenation(input_files=fastqs, output_file=output_file)
    return output_file


def concatenate_reverse_reads(directory: Path) -> Path | None:
    fastqs: list[Path] = get_reverse_read_fastqs(directory)
    if not fastqs:
        return
    file: Path = get_new_unique_file(directory)
    concatenate(input_files=fastqs, output_file=file)
    validate_concatenation(input_files=fastqs, output_file=file)
    return file


def get_new_unique_file(directory: Path) -> Path:
    unique_id = uuid.uuid4()
    return Path(directory, f"{unique_id}.fastq.gz")


def get_forward_read_fastqs(fastq_directory: Path) -> list[Path]:
    return get_fastqs_by_direction(fastq_directory=fastq_directory, direction=1)


def get_reverse_read_fastqs(fastq_directory: Path) -> list[Path]:
    return get_fastqs_by_direction(fastq_directory=fastq_directory, direction=2)


def get_fastqs_by_direction(fastq_directory: Path, direction: int) -> list[Path]:
    pattern = f".+_R{direction}_[0-9]+.fastq.gz"
    fastqs: list[Path] = []
    for file in fastq_directory.iterdir():
        if re.match(pattern, file.name):
            fastqs.append(file)
    return sort_files_by_name(fastqs)


def get_total_size(files: list[Path]) -> int:
    return sum(file.stat().st_size for file in files)


def concatenate(input_files: list[Path], output_file: Path) -> None:
    with open(output_file, "wb") as write_file_obj:
        for file in input_files:
            with open(file, "rb") as file_descriptor:
                shutil.copyfileobj(file_descriptor, write_file_obj)


def validate_concatenation(input_files: list[Path], output_file: Path) -> None:
    total_size: int = get_total_size(input_files)
    concatenated_size: int = get_total_size([output_file])
    if total_size != concatenated_size:
        raise ConcatenationError


def sort_files_by_name(files: list[Path]) -> list[Path]:
    return sorted(files, key=lambda file: file.name)


def file_can_be_removed(file: Path, forward_file: Path, reverse_file: Path) -> bool:
    return file.suffix == ".gz" and file != forward_file and file != reverse_file


def remove_raw_fastqs(fastq_directory: Path, forward_file: Path, reverse_file: Path) -> None:
    for file in fastq_directory.iterdir():
        if file_can_be_removed(file=file, forward_file=forward_file, reverse_file=reverse_file):
            file.unlink()



import datetime as dt
import getpass
import logging
import os
import shutil
from pathlib import Path

import pandas as pd
import paramiko
from housekeeper.store.models import Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.constants.constants import SARS_COV_REGEX
from cg.exc import CgError
from cg.models.cg_config import CGConfig
from cg.models.email import EmailInfo
from cg.store.models import Case, Sample
from cg.store.store import Store
from cg.utils.email import send_mail

LOG = logging.getLogger(__name__)


class FOHMUploadAPI:
    def __init__(self, config: CGConfig, dry_run: bool = False, datestr: str | None = None):
        self.config: CGConfig = config
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.lims_api: LimsAPI = config.lims_api
        self.status_db: Store = config.status_db
        self._dry_run: bool = dry_run
        self._current_datestr: str = datestr or str(dt.date.today())
        self._daily_bundle_path = None
        self._daily_rawdata_path = None
        self._daily_report_path = None
        self._cases_to_aggregate: list[str] = []
        self._daily_reports_list: list[Path] = []
        self._daily_pangolin_list: list[Path] = []
        self._reports_dataframe = None
        self._pangolin_dataframe = None
        self._aggregation_dataframe = None

    @property
    def current_datestr(self) -> str:
        if not self._current_datestr:
            self._current_datestr = str(dt.date.today())
        return self._current_datestr

    @property
    def daily_bundle_path(self) -> Path:
        if not self._daily_bundle_path:
            self._daily_bundle_path: Path = Path(
                Path(self.config.mutant.root).parent, "fohm", self.current_datestr
            )
        return self._daily_bundle_path

    @property
    def daily_rawdata_path(self) -> Path:
        if not self._daily_rawdata_path:
            self._daily_rawdata_path: Path = Path(self.daily_bundle_path, "rawdata")
        return self._daily_rawdata_path

    @property
    def daily_report_path(self) -> Path:
        if not self._daily_report_path:
            self._daily_report_path: Path = Path(self.daily_bundle_path, "reports")
        return self._daily_report_path

    @property
    def reports_dataframe(self) -> pd.DataFrame:
        """Dataframe with all 'komplettering' rows from multiple cases"""
        if not isinstance(self._reports_dataframe, pd.DataFrame):
            self._reports_dataframe = self.create_joined_dataframe(
                self.daily_reports_list
            ).sort_values(by=["provnummer"])
            self._reports_dataframe.drop_duplicates(inplace=True, ignore_index=True)
            self._reports_dataframe = self._reports_dataframe[
                self._reports_dataframe["provnummer"].str.contains(SARS_COV_REGEX)
            ].reset_index(drop=True)
        return self._reports_dataframe

    @property
    def pangolin_dataframe(self) -> pd.DataFrame:
        """Dataframe with all pangolin rows from multiple cases"""
        if not isinstance(self._pangolin_dataframe, pd.DataFrame):
            self._pangolin_dataframe = self.create_joined_dataframe(
                self.daily_pangolin_list
            ).sort_values(by=["taxon"])
            self._pangolin_dataframe.drop_duplicates(inplace=True, ignore_index=True)
            self._pangolin_dataframe = self.pangolin_dataframe[
                self._pangolin_dataframe["taxon"].str.contains(SARS_COV_REGEX)
            ].reset_index(drop=True)
        return self._pangolin_dataframe

    @property
    def aggregation_dataframe(self) -> pd.DataFrame:
        """Dataframe with all 'komplettering' rows from multiple cases, and additional rows to be
        used for aggregation."""
        if not isinstance(self._aggregation_dataframe, pd.DataFrame):
            self._aggregation_dataframe = self.reports_dataframe.copy()
        return self._aggregation_dataframe

    @property
    def daily_reports_list(self) -> list[Path]:
        if not self._daily_reports_list:
            for case_id in self._cases_to_aggregate:
                self._daily_reports_list.append(
                    self.housekeeper_api.get_file_from_latest_version(
                        bundle_name=case_id, tags=["komplettering"]
                    ).full_path
                )
        return self._daily_reports_list

    @property
    def daily_pangolin_list(self) -> list[Path]:
        if not self._daily_pangolin_list:
            for case_id in self._cases_to_aggregate:
                self._daily_pangolin_list.append(
                    self.housekeeper_api.get_file_from_latest_version(
                        bundle_name=case_id, tags=["pangolin-typing-fohm"]
                    ).full_path
                )
        return self._daily_pangolin_list

    @property
    def dry_run(self):
        return self._dry_run

    def check_username(self) -> None:
        if self._dry_run:
            return
        if getpass.getuser() == self.config.fohm.valid_uploader:
            return
        raise CgError(
            f"Cannot upload to FOHM as {getpass.getuser()}, please log in as {self.config.fohm.valid_uploader} "
        )

    def set_cases_to_aggregate(self, cases: list) -> None:
        LOG.info(f"Preparing aggregated delivery for {cases}")
        self._cases_to_aggregate = cases

    def create_daily_delivery_folders(self) -> None:
        LOG.info(f"Creating directory: {self.daily_rawdata_path}")
        LOG.info(f"Creating directory: {self.daily_report_path}")
        if self._dry_run:
            return
        self.daily_rawdata_path.mkdir(parents=True, exist_ok=True)
        self.daily_report_path.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def create_joined_dataframe(file_list: list[Path]) -> pd.DataFrame:
        """Creates dataframe with all csv files used in daily delivery"""
        dataframe_list = [pd.read_csv(filename, index_col=None, header=0) for filename in file_list]
        return pd.concat(dataframe_list, axis=0, ignore_index=True)

    def append_metadata_to_aggregation_df(self) -> None:
        """
        Add fields with internal_id and region_lab to dataframe
        """

        self.aggregation_dataframe["internal_id"] = self.aggregation_dataframe["provnummer"].apply(
            lambda x: self.status_db.get_sample_by_name(name=x).internal_id
        )
        self.aggregation_dataframe["region_lab"] = self.aggregation_dataframe["internal_id"].apply(
            lambda x: f"{self.lims_api.get_sample_attribute(lims_id=x, key='region_code').split(' ')[0]}"
            f"_{self.lims_api.get_sample_attribute(lims_id=x, key='lab_code').split(' ')[0]}"
        )

    def link_sample_rawdata_files(self) -> None:
        """Hardlink samples rawdata files to fohm delivery folder."""
        for sample_id in self.aggregation_dataframe["internal_id"]:
            sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=sample_id)
            bundle_name = sample.links[0].case.internal_id
            version_obj: Version = self.housekeeper_api.last_version(bundle=bundle_name)
            files = self.housekeeper_api.files(version=version_obj.id, tags=[sample_id]).all()
            for file in files:
                if self._dry_run:
                    LOG.info(
                        f"Would have copied {file.full_path} to {Path(self.daily_rawdata_path)}"
                    )
                    continue
                shutil.copy(file.full_path, Path(self.daily_rawdata_path))

    def create_pangolin_reports(self) -> None:
        LOG.info("Creating pangolin reports")
        unique_regionlabs = list(self.aggregation_dataframe["region_lab"].unique())
        LOG.info(f"Regions in batch: {unique_regionlabs}")
        for region_lab in unique_regionlabs:
            LOG.info(f"Aggregating data for {region_lab}")
            pangolin_df = self.pangolin_dataframe[
                self.aggregation_dataframe["region_lab"] == region_lab
            ]
            if self._dry_run:
                LOG.info(pangolin_df)
                continue
            pangolin_path = Path(
                self.daily_rawdata_path,
                f"{region_lab}_{self.current_datestr}_pangolin_classification_format4.txt",
            )
            pangolin_df.to_csv(
                pangolin_path,
                sep="\t",
                index=False,
            )
            pangolin_path.chmod(0o0777)

    def create_komplettering_reports(self) -> None:
        LOG.info("Creating komplettering reports")
        unique_regionlabs = list(self.aggregation_dataframe["region_lab"].unique())
        LOG.info(f"Regions in batch: {unique_regionlabs}")
        for region_lab in unique_regionlabs:
            LOG.info(f"Aggregating data for {region_lab}")
            report_df = self.reports_dataframe[
                self.aggregation_dataframe["region_lab"] == region_lab
            ]
            if self._dry_run:
                LOG.info(report_df)
                continue
            report_df.to_csv(
                self.daily_report_path / f"{region_lab}_{self.current_datestr}_komplettering.csv",
                sep=",",
                index=False,
            )

    def send_mail_reports(self) -> None:
        for file in self.daily_report_path.iterdir():
            LOG.info(
                f"Sending {file.name} report file to {self.config.fohm.email_recipient}, dry run: {self._dry_run}"
            )
            if self._dry_run:
                return

            try:
                send_mail(
                    EmailInfo(
                        receiver_email=self.config.fohm.email_recipient,
                        sender_email=self.config.fohm.email_sender,
                        smtp_server=self.config.fohm.email_host,
                        subject=file.name,
                        message=" ",
                        file=file,
                    )
                )
                file.unlink()

            except Exception as ex:
                LOG.error(f"Failed sending email report {file} with error: {ex}")

        if os.listdir(self.daily_report_path) == []:
            self.daily_report_path.rmdir()

        if os.listdir(self.daily_bundle_path) == []:
            self.daily_bundle_path.rmdir()

    def sync_files_sftp(self) -> None:
        self.check_username()
        transport = paramiko.Transport((self.config.fohm.host, self.config.fohm.port))
        ed_key = paramiko.Ed25519Key.from_private_key_file(self.config.fohm.key)
        transport.connect(username=self.config.fohm.username, pkey=ed_key)
        sftp = paramiko.SFTPClient.from_transport(transport)
        for file in self.daily_rawdata_path.iterdir():
            LOG.info(f"Sending {file} via SFTP, dry-run {self.dry_run}")
            if self._dry_run:
                continue

            try:
                sftp.put(file.as_posix(), f"/till-fohm/{file.name}")
                LOG.info(f"Finished sending {file}")
                file.unlink()
            except Exception as ex:
                LOG.error(f"Failed sending {file} with error: {ex}")

        sftp.close()
        transport.close()

        if os.listdir(self.daily_rawdata_path) == []:
            self.daily_rawdata_path.rmdir()

    def update_upload_started_at(self, case_id: str) -> None:
        """Update timestamp for cases which started being processed as batch"""
        if self._dry_run:
            return
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        case_obj.analyses[0].upload_started_at = dt.datetime.now()
        self.status_db.session.commit()

    def update_uploaded_at(self, case_id: str) -> None:
        """Update timestamp for cases which uploaded successfully"""
        if self._dry_run:
            return
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        case_obj.analyses[0].uploaded_at = dt.datetime.now()
        self.status_db.session.commit()



"""Tests for the DownsampleAPI."""

import logging
from pathlib import Path

import pytest

from cg.apps.downsample.downsample import DownsampleAPI
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI
from cg.models.cg_config import CGConfig
from cg.models.downsample.downsample_data import DownsampleData
from cg.store.models import Sample


def test_add_downsampled_case_entry_to_statusdb(
    downsample_api: DownsampleAPI, downsample_data: DownsampleData, caplog
) -> None:
    """Test to add the downsampled case entry to StatusDB."""
    # GIVEN a DownsampleAPI
    caplog.set_level(level=logging.INFO)
    # WHEN adding a downsampled case to statusDB
    downsample_api.store_downsampled_case(downsample_data)

    # THEN a downsampled case is added to the store
    assert downsample_api.status_db.get_case_by_name(downsample_data.downsampled_case.name)

    # WHEN adding the downsampled case again
    downsample_api.store_downsampled_case(downsample_data)
    # THEN a log info is added
    assert (
        f"Case with name {downsample_data.downsampled_case.name} already exists in StatusDB."
        in caplog.text
    )


def test_add_downsampled_sample_entry_to_statusdb(
    downsample_api: DownsampleAPI, downsample_data: DownsampleData
) -> None:
    """Test to add the downsampled sample entry to StatusDB."""
    # GIVEN a DownsampleAPI

    # WHEN adding a downsampled sample to statusDB
    downsample_api.store_downsampled_sample(downsample_data)

    # THEN a downsampled sample is added to the store
    assert downsample_api.status_db.get_sample_by_internal_id(
        downsample_data.downsampled_sample.internal_id
    )

    # WHEN adding the sample again

    # THEN a ValueError is raised
    with pytest.raises(ValueError):
        downsample_api.store_downsampled_sample(downsample_data)


def test_add_sample_case_links(downsample_api: DownsampleAPI, downsample_data: DownsampleData):
    """Test to link samples to cases in StatusDB."""

    # GIVEN a DownsampleAPI and a downsampled case and sample in StatusDB
    downsample_api.store_downsampled_case(downsample_data)
    downsample_api.store_downsampled_sample(downsample_data)

    # GIVEN the links are not established
    sample: Sample = downsample_api.status_db.get_sample_by_internal_id(
        downsample_data.downsampled_sample.internal_id
    )
    assert not sample.links

    # WHEN adding the sample case links
    downsample_api._link_downsampled_sample_to_case(
        downsample_data=downsample_data,
        sample=downsample_data.downsampled_sample,
        case=downsample_data.downsampled_case,
    )

    # THEN the links are established
    sample: Sample = downsample_api.status_db.get_sample_by_internal_id(
        downsample_data.downsampled_sample.internal_id
    )
    assert sample.links


def test_downsample_api_adding_a_second_sample_to_case(
    downsample_api: DownsampleAPI,
    downsample_data: DownsampleData,
    downsample_sample_internal_id_2: str,
    downsample_case_internal_id: str,
    downsample_case_name: str,
    downsample_context: CGConfig,
):
    """Test that subsequent samples are added to the given case."""
    # GIVEN a DownsampleAPI with a sample and case added
    downsample_api.store_downsampled_sample_case(downsample_data)
    downsample_context.status_db_ = downsample_api.status_db
    downsample_context.housekeeper_api_ = downsample_api.housekeeper_api

    # WHEN generating a new DownsampelData for the same case with a different sample
    new_downsample_data = DownsampleData(
        status_db=downsample_context.status_db_,
        hk_api=downsample_context.housekeeper_api_,
        sample_id=downsample_sample_internal_id_2,
        case_id=downsample_case_internal_id,
        case_name=downsample_case_name,
        number_of_reads=50,
        out_dir=Path(downsample_context.downsample_dir),
    )

    # THEN adding the sample to statusDB adds the sample and generates the links
    downsample_api.store_downsampled_sample_case(new_downsample_data)

    assert new_downsample_data.status_db.get_sample_by_internal_id(
        new_downsample_data.downsampled_sample.internal_id
    )
    assert new_downsample_data.downsampled_case.name == downsample_data.downsampled_case.name


def test_start_downsample_job(
    downsample_api: DownsampleAPI,
    downsample_sample_internal_id_1: str,
    downsample_case_internal_id: str,
    downsample_case_name: str,
    number_of_reads_in_millions: int,
    mocker,
):
    """Test that a downsample job can be started."""

    # GIVEN a DownsampleAPI
    downsample_api.dry_run = True
    # WHEN starting a downsample job
    mocker.patch.object(PrepareFastqAPI, "is_sample_decompression_needed")
    PrepareFastqAPI.is_sample_decompression_needed.return_value = False
    submitted_job: int = downsample_api.downsample_sample(
        sample_id=downsample_sample_internal_id_1,
        case_id=downsample_case_internal_id,
        case_name=downsample_case_name,
        number_of_reads=number_of_reads_in_millions,
    )

    # THEN a job is submitted
    assert submitted_job



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import get_scout_link
from cg.store.models import Case


class ScoutMessage(DeliveryMessage):
    def create_message(self, case: Case) -> str:
        scout_link: str = get_scout_link(case)
        return f"Hello,\n\n The following case has been uploaded to Scout:\n\n {scout_link}"



"""Mock the os ticket api"""

import logging
import os.path

from flask import Flask

from cg.apps.osticket import OsTicket
from cg.exc import TicketCreationError

LOG = logging.getLogger(__name__)


class MockOsTicket(OsTicket):
    """Interface to ticket system."""

    def __init__(self):
        self.headers = None
        self.url = None
        self.osticket_email = "james.holden@scilifelab.se"
        self.mail_container_uri = "dummy_uri"
        self._ticket_nr: str = "123456"
        self._should_fail: bool = False
        self._return_none: bool = False
        self.email_uri = "http://localhost:0000/sendmail"

    def set_ticket_nr(self, ticket_id: str) -> None:
        self._ticket_nr = ticket_id

    def init_app(self, app: Flask):
        """Initialize the API in Flask."""

    def setup(
        self,
        api_key: str = None,
        domain: str = None,
        osticket_email: str = None,
        email_uri: str = None,
    ):
        """Initialize the API."""
        self.headers = {"X-API-Key": api_key}
        self.url = os.path.join(domain, "api/tickets.json")

    def open_ticket(
        self, attachment: dict, email: str, message: str, name: str, subject: str
    ) -> str | None:
        """Open a new ticket through the REST API."""
        if self._should_fail:
            LOG.error("res.text: %s, reason: %s", self._ticket_nr, "Unknown reason")
            raise TicketCreationError("FAIL")
        if self._return_none:
            return None

        return self._ticket_nr



import logging

import mock
from click.testing import CliRunner, Result

from cg.cli.upload.scout import create_scout_load_config, get_upload_api
from cg.constants import EXIT_SUCCESS, Workflow
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.meta.upload.upload_api import UploadAPI
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case
from cg.store.store import Store
from tests.mocks.scout import MockScoutLoadConfig
from tests.store_helpers import StoreHelpers


def test_get_upload_api(cg_context: CGConfig, case_id: str, helpers: StoreHelpers):
    """Test to get the correct upload API for a BALSAMIC case."""
    status_db: Store = cg_context.status_db

    # GIVEN a case with a balsamic analysis
    case: Case = helpers.ensure_case(
        store=status_db, data_analysis=Workflow.BALSAMIC, case_id=case_id
    )
    helpers.add_analysis(store=status_db, case=case, workflow=Workflow.BALSAMIC)

    # WHEN getting the upload API
    upload_api: UploadAPI = get_upload_api(cg_config=cg_context, case=case)

    # THEN assert that the type of upload API is correct
    assert type(upload_api.analysis_api) == BalsamicAnalysisAPI


def test_create_scout_load_config(
    caplog,
    cli_runner: CliRunner,
    cg_context: CGConfig,
    case_id: str,
    helpers: StoreHelpers,
):
    """Test to create a scout load config for a BALSAMIC case."""
    caplog.set_level(logging.DEBUG)
    status_db: Store = cg_context.status_db

    # GIVEN a case with a balsamic analysis
    case: Case = helpers.ensure_case(
        store=status_db, data_analysis=Workflow.BALSAMIC, case_id=case_id
    )
    helpers.add_analysis(store=status_db, case=case, workflow=Workflow.BALSAMIC)

    with mock.patch.object(UploadScoutAPI, "generate_config", return_value=MockScoutLoadConfig()):
        # WHEN creating the scout load config
        result: Result = cli_runner.invoke(
            create_scout_load_config, ["--print", case_id], obj=cg_context
        )

        # THEN assert that the code exits with success
        assert result.exit_code == EXIT_SUCCESS
        # THEN assert that the root path is in the log
        assert cg_context.meta_apis["upload_api"].analysis_api.root in caplog.text



"""Code for uploading genotype data via CLI"""

import logging

import click

from cg.meta.upload.gisaid import GisaidAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.command("gisaid")
@click.argument("case_id", required=True)
@click.pass_obj
def upload_to_gisaid(context: CGConfig, case_id: str):
    """Upload mutant analysis data to GISAID."""

    LOG.info("----------------- GISAID UPLOAD -------------------")

    gisaid_api = GisaidAPI(config=context)

    gisaid_api.upload(case_id=case_id)
    LOG.info("Upload to GISAID successful")



import pytest
from cg.apps.tb.api import TrailblazerAPI
from cg.apps.tb.dto.summary_response import AnalysisSummary

from cg.services.orders.order_status_service.order_status_service import OrderStatusService
from cg.store.models import Case, Customer, Order
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture
def order_status_service(trailblazer_api: TrailblazerAPI, store: Store):
    return OrderStatusService(analysis_client=trailblazer_api, store=store)


@pytest.fixture
def order(store: Store, helpers: StoreHelpers) -> Order:
    customer: Customer = helpers.ensure_customer(store)
    order: Order = helpers.add_order(store=store, customer_id=customer.id, ticket_id="ticket_id")
    helpers.ensure_case(store=store, customer=customer, order=order)
    return order


@pytest.fixture
def analysis_summary(order: Order) -> AnalysisSummary:
    return AnalysisSummary(
        order_id=order.id,
        analysis_id=1,
        delivered=2,
        running=3,
        cancelled=4,
        failed=5,
    )



"""Tests for the calculations module."""

from cg.utils.calculations import multiply_by_million


def test_multiple_by_a_million():
    # GIVEN a number
    number: float = 1.00

    # WHEN multiplying by a million
    multiplied: int = multiply_by_million(number)

    # THEN the number should be multiplied by a million
    assert multiplied == int(number * 1_000_000)



from cg.store.models import Flowcell
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_update_flow_cell_has_backup(base_store: Store, flow_cell_name: str, helpers: StoreHelpers):
    """Test updating the backup status of a flow cell in the database."""

    # GIVEN a database containing a flow cell without a back-up
    flow_cell: Flowcell = helpers.add_flow_cell(
        store=base_store, flow_cell_name=flow_cell_name, has_backup=False
    )

    assert not flow_cell.has_backup

    # WHEN updating flow cell attribute has back-up
    base_store.update_flow_cell_has_backup(flow_cell=flow_cell, has_backup=True)

    # THEN flow cell has backup should be true
    assert flow_cell.has_backup



"""Tests for the collect qc metrics module."""

from housekeeper.store.models import File

from cg.clients.janus.dto.create_qc_metrics_request import CreateQCMetricsRequest
from cg.meta.qc_metrics.collect_qc_metrics import CollectQCMetricsAPI


def test_get_qc_metrics_file_paths_for_case(
    case_id_with_single_sample: str,
    collect_qc_metrics_api: CollectQCMetricsAPI,
    janus_hk_tags: list[str],
    file_with_right_tags: str,
):
    # GIVEN a Housekeeper store that contains a case with a qc metrics file

    # WHEN retrieving the files
    files: list[File] = collect_qc_metrics_api.get_qc_metrics_file_paths_for_case(
        case_id_with_single_sample
    )

    # THEN file are returned
    assert files
    assert len(files) == 1


def test_create_qc_metrics_request(
    collect_qc_metrics_api: CollectQCMetricsAPI,
    case_id_with_single_sample: str,
    expected_request: CreateQCMetricsRequest,
):
    # GIVEN a collect qc metrics api

    # WHEN creating a qc metrics request for a case id
    qc_metrics_request: CreateQCMetricsRequest = collect_qc_metrics_api.create_qc_metrics_request(
        case_id_with_single_sample
    )

    # THEN a qc metrics request is created
    assert qc_metrics_request == expected_request



from pathlib import Path

import pytest

from cg.meta.workflow.fastq import FastqHandler
from cg.models.fastq import FastqFileMeta


@pytest.mark.parametrize(
    "fastq_header, expected_header_meta",
    [
        (
            "@HWUSI-EAS100R:6:73:941:1973#0/1",
            FastqFileMeta(lane=6, flow_cell_id="XXXXXX", read_direction=1),
        ),
        (
            "@EAS139:136:FC706VJ:2:2104:15343:197393 1:Y:18:ATCACG",
            FastqFileMeta(lane=2, flow_cell_id="FC706VJ", read_direction=1),
        ),
        (
            "@ST-E00201:173:HCLCGALXX:1:2106:22516:34834/1",
            FastqFileMeta(lane=1, flow_cell_id="HCLCGALXX", read_direction=1),
        ),
    ],
)
def test_parse_fastq_header(fastq_header: str, expected_header_meta: dict):
    # GIVEN a FASTQ header

    # WHEN parsing header
    header_meta: FastqFileMeta = FastqHandler.parse_fastq_header(line=fastq_header)

    # THEN header meta should match the expected header information
    assert header_meta == expected_header_meta


@pytest.mark.parametrize(
    "fastq_header, expected_error",
    [
        ("no match", TypeError),
        ("no:match", TypeError),
        ("", TypeError),
    ],
)
def test_parse_fastq_header_when_no_match(fastq_header: str, expected_error):
    # GIVEN no matching FASTQ header

    with pytest.raises(expected_error):
        # WHEN parsing header
        # THEN raise error
        FastqHandler.parse_fastq_header(line=fastq_header)


@pytest.mark.parametrize(
    "fastq_path, expected_fastq_meta",
    [
        (
            Path("tests", "fixtures", "io", "casava_five_parts.fastq.gz"),
            FastqFileMeta(
                path=Path("tests", "fixtures", "io", "casava_five_parts.fastq.gz"),
                lane=6,
                flow_cell_id="XXXXXX",
                read_direction=1,
                undetermined=None,
            ),
        ),
        (
            Path("tests", "fixtures", "io", "casava_ten_parts.fastq.gz"),
            FastqFileMeta(
                path=Path("tests", "fixtures", "io", "casava_ten_parts.fastq.gz"),
                lane=2,
                flow_cell_id="FC706VJ",
                read_direction=1,
                undetermined=None,
            ),
        ),
        (
            Path("tests", "fixtures", "io", "casava_seven_parts.fastq.gz"),
            FastqFileMeta(
                path=Path("tests", "fixtures", "io", "casava_seven_parts.fastq.gz"),
                lane=1,
                flow_cell_id="HCLCGALXX",
                read_direction=1,
                undetermined=None,
            ),
        ),
    ],
)
def test_parse_file_data(fastq_path: Path, expected_fastq_meta: dict, mocker):
    # GIVEN a FASTQ file

    with mocker.patch("cg.meta.workflow.fastq._is_undetermined_in_path", return_value=None):
        # WHEN parsing header
        header_meta = FastqHandler.parse_file_data(fastq_path=fastq_path)

        # THEN header meta should match the expected header information
        assert header_meta == expected_fastq_meta



import datetime as dt

from cg.constants import DataDelivery, GenePanelMasterList
from cg.constants.constants import CustomerId, PrepCategory, Workflow
from cg.constants.priority import Priority
from cg.exc import OrderError
from cg.meta.orders.lims import process_lims
from cg.meta.orders.submitter import Submitter
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import StatusEnum
from cg.store.models import ApplicationVersion, Case, CaseSample, Customer, Sample


class FastqSubmitter(Submitter):
    def submit_order(self, order: OrderIn) -> dict:
        """Submit a batch of samples for FASTQ delivery."""

        project_data, lims_map = process_lims(
            lims_api=self.lims, lims_order=order, new_samples=order.samples
        )
        status_data = self.order_to_status(order)
        self._fill_in_sample_ids(samples=status_data["samples"], lims_map=lims_map)
        new_samples = self.store_items_in_status(
            customer_id=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"],
            ticket_id=order.ticket,
            items=status_data["samples"],
        )
        self._add_missing_reads(new_samples)
        return {"project": project_data, "records": new_samples}

    @staticmethod
    def order_to_status(order: OrderIn) -> dict:
        """Convert order input to status for fastq-only orders."""
        status_data = {
            "customer": order.customer,
            "order": order.name,
            "samples": [
                {
                    "application": sample.application,
                    "capture_kit": sample.capture_kit,
                    "comment": sample.comment,
                    "data_analysis": sample.data_analysis,
                    "data_delivery": sample.data_delivery,
                    "name": sample.name,
                    "priority": sample.priority,
                    "sex": sample.sex,
                    "subject_id": sample.subject_id,
                    "tumour": sample.tumour,
                    "volume": sample.volume,
                }
                for sample in order.samples
            ],
        }
        return status_data

    def create_maf_case(self, sample_obj: Sample) -> None:
        """Add a MAF case to the Status database."""
        case: Case = self.status.add_case(
            data_analysis=Workflow(Workflow.MIP_DNA),
            data_delivery=DataDelivery(DataDelivery.NO_DELIVERY),
            name="_".join([sample_obj.name, "MAF"]),
            panels=[GenePanelMasterList.OMIM_AUTO],
            priority=Priority.research,
            ticket=sample_obj.original_ticket,
        )
        case.customer = self.status.get_customer_by_internal_id(
            customer_internal_id=CustomerId.CG_INTERNAL_CUSTOMER
        )
        relationship: CaseSample = self.status.relate_sample(
            case=case, sample=sample_obj, status=StatusEnum.unknown
        )
        self.status.session.add_all([case, relationship])

    def store_items_in_status(
        self, customer_id: str, order: str, ordered: dt.datetime, ticket_id: str, items: list[dict]
    ) -> list[Sample]:
        """Store fastq samples in the status database including family connection and delivery"""
        customer: Customer = self.status.get_customer_by_internal_id(
            customer_internal_id=customer_id
        )
        if not customer:
            raise OrderError(f"Unknown customer: {customer_id}")
        new_samples = []
        case: Case = self.status.get_case_by_name_and_customer(
            customer=customer, case_name=ticket_id
        )
        submitted_case: dict = items[0]
        with self.status.session.no_autoflush:
            for sample in items:
                new_sample = self.status.add_sample(
                    name=sample["name"],
                    sex=sample["sex"] or "unknown",
                    comment=sample["comment"],
                    internal_id=sample.get("internal_id"),
                    order=order,
                    ordered=ordered,
                    original_ticket=ticket_id,
                    priority=sample["priority"],
                    tumour=sample["tumour"],
                    capture_kit=sample["capture_kit"],
                    subject_id=sample["subject_id"],
                )
                new_sample.customer: Customer = customer
                application_tag: str = sample["application"]
                application_version: ApplicationVersion = (
                    self.status.get_current_application_version_by_tag(tag=application_tag)
                )
                if application_version is None:
                    raise OrderError(f"Invalid application: {sample['application']}")
                new_sample.application_version: ApplicationVersion = application_version
                new_samples.append(new_sample)
                if not case:
                    case = self.status.add_case(
                        data_analysis=Workflow(submitted_case["data_analysis"]),
                        data_delivery=DataDelivery(submitted_case["data_delivery"]),
                        name=ticket_id,
                        panels=None,
                        priority=submitted_case["priority"],
                        ticket=ticket_id,
                    )
                if (
                    not new_sample.is_tumour
                    and new_sample.prep_category == PrepCategory.WHOLE_GENOME_SEQUENCING
                ):
                    self.create_maf_case(sample_obj=new_sample)
                case.customer = customer
                new_relationship = self.status.relate_sample(
                    case=case, sample=new_sample, status=StatusEnum.unknown
                )
                new_delivery = self.status.add_delivery(destination="caesar", sample=new_sample)
                self.status.session.add_all([case, new_relationship, new_delivery])

        self.status.session.add_all(new_samples)
        self.status.session.commit()
        return new_samples



"""Module for Balsamic QC Analysis API."""

import logging

from cg.constants import Workflow
from cg.meta.workflow.balsamic import BalsamicAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


class BalsamicQCAnalysisAPI(BalsamicAnalysisAPI):
    """Handles communication between BALSAMIC processes
    and the rest of CG infrastructure"""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.BALSAMIC_QC,
    ):
        super().__init__(config=config, workflow=workflow)



from click.testing import CliRunner

from cg.cli.add import add
from cg.constants import EXIT_SUCCESS, Priority
from cg.constants.subject import Sex
from cg.models.cg_config import CGConfig
from cg.store.models import Customer, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_add_sample_missing_customer(cli_runner: CliRunner, base_context: CGConfig):
    """Test adding a sample when customer id supplied does not match a customer in the database."""
    # GIVEN an empty database
    disk_store: Store = base_context.status_db

    # WHEN adding a sample
    application = "dummy_application"
    customer_id = "dummy_customer"
    name = "dummy_name"
    result = cli_runner.invoke(
        add,
        [
            "sample",
            "--sex",
            Sex.MALE,
            "--application-tag",
            application,
            customer_id,
            name,
        ],
        obj=base_context,
    )

    # THEN it should complain about missing customer instead of adding a sample
    assert result.exit_code == 1
    assert disk_store._get_query(table=Sample).count() == 0


def test_add_sample_bad_application(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test adding a sample when application tag supplied does not match an application tag in the database."""
    # GIVEN a database with a customer
    disk_store: Store = base_context.status_db

    # WHEN adding a sample
    application = "dummy_application"
    customer: Customer = helpers.ensure_customer(store=disk_store)
    name = "dummy_name"
    result = cli_runner.invoke(
        add,
        [
            "sample",
            "--sex",
            Sex.MALE,
            "--application-tag",
            application,
            customer.internal_id,
            name,
        ],
        obj=base_context,
    )

    # THEN it should complain about missing application instead of adding a sample
    assert result.exit_code == 1
    assert disk_store._get_query(table=Sample).count() == 0


def test_add_sample_required(cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers):
    """Test adding a sample."""
    # GIVEN a database with a customer and an application
    disk_store: Store = base_context.status_db
    application_tag = "dummy_tag"
    helpers.ensure_application(store=disk_store, tag=application_tag)
    helpers.ensure_application_version(store=disk_store, application_tag=application_tag)
    customer: Customer = helpers.ensure_customer(store=disk_store)
    name = "sample_name"

    # WHEN adding a sample
    result = cli_runner.invoke(
        add,
        [
            "sample",
            "--sex",
            Sex.MALE,
            "--application-tag",
            application_tag,
            customer.internal_id,
            name,
        ],
        obj=base_context,
    )

    # THEN it should be added
    assert result.exit_code == EXIT_SUCCESS
    sample_query = disk_store._get_query(table=Sample)
    assert sample_query.count() == 1
    assert sample_query.first().name == name
    assert sample_query.first().sex == Sex.MALE


def test_add_sample_lims_id(cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers):
    """Test adding a sample using a LIMS id."""
    # GIVEN a database with a customer and an application
    disk_store: Store = base_context.status_db
    application_tag = "dummy_tag"
    helpers.ensure_application(store=disk_store, tag=application_tag)
    helpers.ensure_application_version(store=disk_store, application_tag=application_tag)
    # WHEN adding a sample
    customer: Customer = helpers.ensure_customer(store=disk_store)
    name = "sample_name"
    lims_id = "sample_lims_id"

    result = cli_runner.invoke(
        add,
        [
            "sample",
            "--sex",
            Sex.MALE,
            "--application-tag",
            application_tag,
            "--lims",
            lims_id,
            customer.internal_id,
            name,
        ],
        obj=base_context,
    )

    # THEN it should be added
    assert result.exit_code == EXIT_SUCCESS
    sample_query = disk_store._get_query(table=Sample)
    assert sample_query.count() == 1
    assert sample_query.first().internal_id == lims_id


def test_add_sample_order(
    cli_runner: CliRunner,
    base_context: CGConfig,
    disk_store: Store,
    helpers: StoreHelpers,
    application_tag: str,
):
    """Test adding a sample using an external sample id."""
    # GIVEN a database with a customer and an application
    helpers.ensure_application(store=disk_store, tag=application_tag)
    helpers.ensure_application_version(store=disk_store, application_tag=application_tag)
    customer: Customer = helpers.ensure_customer(store=disk_store)
    name = "sample_name"
    order = "sample_order"

    # WHEN adding a sample
    result = cli_runner.invoke(
        add,
        [
            "sample",
            "--sex",
            Sex.MALE,
            "--application-tag",
            application_tag,
            "--order",
            order,
            customer.internal_id,
            name,
        ],
        obj=base_context,
    )

    # THEN it should be added
    assert result.exit_code == EXIT_SUCCESS
    sample_query = disk_store._get_query(table=Sample)
    assert sample_query.count() == 1
    assert sample_query.first().order == order


def test_add_sample_when_down_sampled(
    cli_runner,
    base_context: CGConfig,
    disk_store: Store,
    application_tag: str,
    helpers: StoreHelpers,
):
    """Test adding a sample when down sampled"""
    # GIVEN a database with a customer and an application
    helpers.ensure_application(store=disk_store, tag=application_tag)
    helpers.ensure_application_version(store=disk_store, application_tag=application_tag)
    customer: Customer = helpers.ensure_customer(store=disk_store)
    name = "sample_name"
    down_sampled_to = "123"

    # WHEN adding a sample
    result = cli_runner.invoke(
        add,
        [
            "sample",
            "--sex",
            Sex.MALE,
            "--application-tag",
            application_tag,
            "--down-sampled",
            down_sampled_to,
            customer.internal_id,
            name,
        ],
        obj=base_context,
    )

    # THEN it should be added
    assert result.exit_code == EXIT_SUCCESS
    sample_query = disk_store._get_query(table=Sample)
    assert sample_query.count() == 1
    assert str(sample_query.first().downsampled_to) == down_sampled_to


def test_add_sample_priority(
    cli_runner: CliRunner,
    base_context: CGConfig,
    disk_store: Store,
    application_tag: str,
    helpers: StoreHelpers,
):
    """Test adding a sample with priority."""
    # GIVEN a database with a customer and an application
    helpers.ensure_application(store=disk_store, tag=application_tag)
    helpers.ensure_application_version(store=disk_store, application_tag=application_tag)
    customer: Customer = helpers.ensure_customer(store=disk_store)
    name = "sample_name"

    # WHEN adding a sample
    result = cli_runner.invoke(
        add,
        [
            "sample",
            "--sex",
            Sex.MALE,
            "--application-tag",
            application_tag,
            "--priority",
            Priority.priority.name,
            customer.internal_id,
            name,
        ],
        obj=base_context,
    )

    # THEN it should be added
    assert result.exit_code == EXIT_SUCCESS
    sample_query = disk_store._get_query(table=Sample)
    assert sample_query.count() == 1
    assert sample_query.first().priority_human == Priority.priority.name



import re

from pydantic import AfterValidator
from typing_extensions import Annotated


def is_valid_sample_internal_id(sample_internal_id: str) -> bool:
    """
    Check if a sample internal id has the correct structure:
    starts with three letters followed by at least three digits.
    """
    return bool(re.search(r"^[A-Za-z]{3}\d{3}", sample_internal_id))


def validate_sample_id(sample_id: str) -> str:
    """Validate a sample id."""
    if not is_valid_sample_internal_id(sample_id):
        raise ValueError(f"{sample_id} is not a valid sample id")
    return sample_id


def remove_index_from_sample_id(sample_id_with_index: str) -> str:
    return sample_id_with_index.split("_")[0]


SampleId = Annotated[
    str,
    AfterValidator(remove_index_from_sample_id),
    AfterValidator(validate_sample_id),
]



"""Module for Raredisease Analysis API."""

import logging
from pathlib import Path

from cg.constants import GenePanelMasterList, Workflow
from cg.constants.gene_panel import GENOME_BUILD_37
from cg.meta.workflow.analysis import add_gene_panel_combo
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


class RarediseaseAnalysisAPI(NfAnalysisAPI):
    """Handles communication between RAREDISEASE processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.RAREDISEASE,
    ):
        super().__init__(config=config, workflow=workflow)

    @property
    def root(self) -> str:
        return self.config.raredisease.root

    def write_managed_variants(self, case_id: str, content: list[str]) -> None:
        self._write_managed_variants(out_dir=Path(self.root, case_id), content=content)

    def write_panel(self, case_id: str, content: list[str]) -> None:
        """Write the gene panel to case dir."""
        self._write_panel(out_dir=Path(self.root, case_id), content=content)

    @staticmethod
    def get_aggregated_panels(customer_id: str, default_panels: set[str]) -> list[str]:
        """Check if customer should use the gene panel master list
        and if all default panels are included in the gene panel master list.
        If not, add gene panel combo and OMIM-AUTO.
        Return an aggregated gene panel."""
        master_list: list[str] = GenePanelMasterList.get_panel_names()
        if customer_id in GenePanelMasterList.collaborators() and default_panels.issubset(
            master_list
        ):
            return master_list
        all_panels: set[str] = add_gene_panel_combo(default_panels=default_panels)
        all_panels |= {GenePanelMasterList.OMIM_AUTO, GenePanelMasterList.PANELAPP_GREEN}
        return list(all_panels)

    def get_gene_panel(self, case_id: str) -> list[str]:
        """Create and return the aggregated gene panel file."""
        return self._get_gene_panel(case_id=case_id, genome_build=GENOME_BUILD_37)

    def get_managed_variants(self) -> list[str]:
        """Create and return the managed variants."""
        return self._get_managed_variants(genome_build=GENOME_BUILD_37)



from pydantic import BaseModel

from cg.models.orders.sample_base import OrderSample


# Class for holding information about cases in order
class OrderCase(BaseModel):
    cohorts: list[str] | None
    name: str
    panels: list[str] | None
    priority: str
    samples: list[OrderSample]
    synopsis: str | None


class OrderPool(BaseModel):
    name: str
    data_analysis: str
    data_delivery: str | None
    application: str
    samples: list[OrderSample]


# This is for validating in data
class Orderform(BaseModel):
    comment: str | None = None
    delivery_type: str
    project_type: str
    customer: str
    name: str
    data_analysis: str | None = None
    data_delivery: str | None = None
    ticket: int | None = None
    samples: list[OrderSample]
    cases: list[OrderCase] | None = None



from pathlib import Path

import pkg_resources

VALID_INDEXES_FILE: str = Path("resources", "20181012_Indices.csv").as_posix()

VALID_INDEXES_PATH: Path = Path(pkg_resources.resource_filename("cg", VALID_INDEXES_FILE))

RNAFUSION_BUNDLE_FILENAMES: str = Path("resources", "rnafusion_bundle_filenames.csv").as_posix()

TAXPROFILER_BUNDLE_FILENAMES: str = Path("resources", "taxprofiler_bundle_filenames.csv").as_posix()

RNAFUSION_BUNDLE_FILENAMES_PATH: Path = Path(
    pkg_resources.resource_filename("cg", RNAFUSION_BUNDLE_FILENAMES)
)

TAXPROFILER_BUNDLE_FILENAMES_PATH: Path = Path(
    pkg_resources.resource_filename("cg", TAXPROFILER_BUNDLE_FILENAMES)
)



"""Module for the collect qc metrics meta api."""

import logging
from typing import Iterator

from housekeeper.store.models import File, Tag

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.clients.arnold.api import ArnoldAPIClient
from cg.clients.janus.api import JanusAPIClient
from cg.clients.janus.dto.create_qc_metrics_request import (
    CreateQCMetricsRequest,
    FilePathAndTag,
    WorkflowInfo,
)
from cg.clients.janus.exceptions import JanusClientError, JanusServerError
from cg.constants.housekeeper_tags import JanusTags
from cg.exc import HousekeeperFileMissingError
from cg.store.models import Case
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class CollectQCMetricsAPI:
    """Collect qc metrics API."""

    def __init__(
        self,
        hk_api: HousekeeperAPI,
        status_db: Store,
        janus_api: JanusAPIClient,
        arnold_api: ArnoldAPIClient,
    ):
        self.hk_api: HousekeeperAPI = hk_api
        self.status_db: Store = status_db
        self.janus_api: JanusAPIClient = janus_api
        self.arnold_api: ArnoldAPIClient = arnold_api

    def get_qc_metrics_file_paths_for_case(self, case_id: str) -> list[File]:
        """Get the multiqc metrics files for a case."""
        files: list[File] = self.hk_api.get_files(
            bundle=case_id, tags=JanusTags.tags_to_retrieve
        ).all()
        if not files:
            raise HousekeeperFileMissingError(f"No qc metrics files found for {case_id}.")
        return files

    @staticmethod
    def get_tag_to_include(tags: list[Tag]) -> str:
        """Get the multiqc module file tag."""
        for tag in tags:
            if tag.name in JanusTags.multi_qc_file_tags:
                return tag.name

    def get_file_paths_and_tags(self, files: list[File]) -> list[FilePathAndTag]:
        """Get the file path and tag for the multiqc files."""
        file_paths_and_tags: list[FilePathAndTag] = []
        for hk_file in files:
            file_path: str = hk_file.full_path
            tag: str = self.get_tag_to_include(hk_file.tags)
            file_paths_and_tags.append(FilePathAndTag(file_path=file_path, tag=tag))
        return file_paths_and_tags

    @staticmethod
    def get_workflow_info(case: Case) -> WorkflowInfo:
        workflow: str = case.data_analysis
        workflow_version: str = case.analyses[0].workflow_version
        return WorkflowInfo(workflow=workflow, version=workflow_version)

    def create_qc_metrics_request(self, case_id: str) -> CreateQCMetricsRequest:
        """Create a qc metrics request."""
        files: list[File] = self.get_qc_metrics_file_paths_for_case(case_id)
        file_paths_and_tags: list[FilePathAndTag] = self.get_file_paths_and_tags(files)
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        workflow_info: WorkflowInfo = self.get_workflow_info(case)
        sample_ids: Iterator[str] = self.status_db.get_sample_ids_by_case_id(case_id)

        return CreateQCMetricsRequest(
            case_id=case_id,
            sample_ids=sample_ids,
            workflow_info=workflow_info,
            files=file_paths_and_tags,
        )

    def get_qc_metrics(self, case_id) -> dict:
        """Get the qc metrics for a case."""
        qc_metrics_request: CreateQCMetricsRequest = self.create_qc_metrics_request(case_id)
        try:
            qc_metrics: dict = self.janus_api.qc_metrics(qc_metrics_request)
            return qc_metrics
        except (JanusClientError, JanusServerError) as error:
            LOG.info(f"Cannot collect qc metrics from Janus: {error}")



import logging
from datetime import datetime, timedelta
from pathlib import Path

from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI

LOG = logging.getLogger(__name__)


class CleanRetrievedSpringFilesAPI:
    """API for cleaning archived Spring files which have been retrieved."""

    def __init__(self, housekeeper_api: HousekeeperAPI, dry_run: bool):
        self.housekeeper_api: HousekeeperAPI = housekeeper_api
        self.dry_run = dry_run

    def _get_files_to_remove(self, age_limit: int) -> list[File]:
        """Returns all Spring files which were retrieved more than given amount of days ago."""
        return self.housekeeper_api.get_spring_files_retrieved_before(
            date=datetime.now() - timedelta(days=age_limit)
        )

    def _unlink_files(self, files_to_unlink: list[File]) -> None:
        for file in files_to_unlink:
            file_path: str = file.full_path
            if self.dry_run:
                LOG.info(f"Dry run - would have unlinked {file_path}")
            LOG.info(f"Unlinking {file_path}")
            Path(file_path).unlink(missing_ok=True)

    def clean_retrieved_spring_files(self, age_limit: int):
        """Removes Spring files retrieved more than given amount of days ago from Hasta,
        and resets retrieval data in Housekeeper."""

        LOG.info("Starting cleaning of retrieved Spring files.")
        files_to_remove: list[File] = self._get_files_to_remove(age_limit)
        if not files_to_remove:
            LOG.info("No retrieved Spring files to remove - exiting.")
            return
        self._unlink_files(files_to_remove)
        if not self.dry_run:
            self.housekeeper_api.reset_retrieved_archive_data(files_to_remove)
        else:
            LOG.info("Would have reset the files' retrieval data")



import pytest
from _pytest.fixtures import FixtureRequest

from cg.apps.demultiplex.sample_sheet.sample_sheet_validator import SampleSheetValidator
from cg.constants.demultiplexing import SampleSheetBCLConvertSections
from cg.exc import SampleSheetError


@pytest.mark.parametrize(
    "sample_sheet_content",
    [
        "hiseq_x_single_index_sample_sheet_content",
        "hiseq_x_dual_index_sample_sheet_content",
        "hiseq_2500_dual_index_sample_sheet_content",
        "hiseq_2500_custom_index_sample_sheet_content",
        "novaseq_6000_pre_1_5_kits_sample_sheet_content",
        "novaseq_6000_post_1_5_kits_sample_sheet_content",
        "novaseq_x_sample_sheet_content",
    ],
    ids=[
        "HiSeqXSingleIndex",
        "HiSeqXDualIndex",
        "HiSeq2500DualIndex",
        "HiSeq2500CustomIndex",
        "NovaSeq6000Pre1.5Kits",
        "NovaSeq6000Post1.5Kits",
        "NovaSeqX",
    ],
)
def test_validate_all_sections_present(
    sample_sheet_validator: SampleSheetValidator,
    sample_sheet_content: str,
    request: FixtureRequest,
):
    """Test that when all sections are present in the sample sheet the validation passes."""
    # GIVEN a sample sheet content with all the required sections and a sample sheet validator
    content: list[list[str]] = request.getfixturevalue(sample_sheet_content)
    sample_sheet_validator.set_sample_sheet_content(content)

    # WHEN validating the sections of the sample sheet
    sample_sheet_validator._validate_all_sections_present()

    # THEN no error is raised


def test_validate_all_sections_present_missing_section(
    sample_sheet_validator: SampleSheetValidator,
    sample_sheet_content_missing_data_header: list[list[str]],
    caplog: pytest.LogCaptureFixture,
):
    """Test that when one sections is missing in the sample sheet the validation fails."""
    # GIVEN a sample sheet content with a missing section and a sample sheet validator
    sample_sheet_validator.set_sample_sheet_content(sample_sheet_content_missing_data_header)
    assert len(sample_sheet_validator.content) == 3

    # WHEN validating the sections of the sample sheet
    with pytest.raises(SampleSheetError):
        # THEN a SampleSheetError is raised
        sample_sheet_validator._validate_all_sections_present()
    assert "Sample sheet does not have all the necessary sections" in caplog.text


@pytest.mark.parametrize(
    "sample_sheet_content, expected_index_settings_name",
    [
        ("novaseq_6000_pre_1_5_kits_sample_sheet_content", "NoReverseComplements"),
        ("novaseq_6000_post_1_5_kits_sample_sheet_content", "NovaSeq6000Post1.5Kits"),
        ("novaseq_x_sample_sheet_content", "NovaSeqX"),
    ],
    ids=["NovaSeq6000Pre1.5Kits", "NovaSeq6000Post1.5Kits", "NovaSeqX"],
)
def test_get_index_settings_name(
    sample_sheet_validator: SampleSheetValidator,
    sample_sheet_content: str,
    expected_index_settings_name: str,
    request: FixtureRequest,
):
    """Test that the correct index settings name is returned."""
    # GIVEN a sample sheet validator with a valid content
    sample_sheet_validator.set_sample_sheet_content(request.getfixturevalue(sample_sheet_content))

    # WHEN getting the index settings name
    index_settings_name: str = sample_sheet_validator._get_index_settings_name()

    # THEN the correct index settings name is returned
    assert index_settings_name == expected_index_settings_name


def test_get_index_settings_name_missing_index_settings(
    sample_sheet_validator: SampleSheetValidator,
    novaseq_x_sample_sheet_content: list[list[str]],
    caplog,
):
    """Test that getting the index settings from a sample sheet without it fails."""
    # GIVEN the content of a sample sheet without index settings
    novaseq_x_sample_sheet_content.pop(5)
    sample_sheet_validator.set_sample_sheet_content(novaseq_x_sample_sheet_content)
    assert [
        SampleSheetBCLConvertSections.Header.INDEX_SETTINGS.value,
        "NovaSeqX",
    ] not in sample_sheet_validator.content

    # WHEN getting the index settings name
    with pytest.raises(SampleSheetError):
        # THEN a SampleSheetError is raised
        sample_sheet_validator._get_index_settings_name()
    assert "No index settings found in sample sheet" in caplog.text


@pytest.mark.parametrize(
    "sample_sheet_content, expected_reverse_complement",
    [
        ("novaseq_6000_pre_1_5_kits_sample_sheet_content", False),
        ("novaseq_6000_post_1_5_kits_sample_sheet_content", False),
        ("novaseq_x_sample_sheet_content", True),
    ],
    ids=["NovaSeq6000Pre1.5Kits", "NovaSeq6000Post1.5Kits", "NovaSeqX"],
)
def test_set_is_index2_reverse_complement(
    sample_sheet_validator: SampleSheetValidator,
    sample_sheet_content: str,
    expected_reverse_complement: bool,
    request: FixtureRequest,
):
    """Test that the correct value for index2 reverse complement value is set."""
    # GIVEN a sample sheet validator with a valid content
    sample_sheet_validator.set_sample_sheet_content(request.getfixturevalue(sample_sheet_content))

    # WHEN setting the index2 reverse complement value
    sample_sheet_validator._set_is_index2_reverse_complement()

    # THEN the correct value is set
    assert sample_sheet_validator.is_index2_reverse_complement == expected_reverse_complement


@pytest.mark.parametrize(
    "sample_sheet_content, nullable, expected",
    [
        ([[SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_2, 10]], False, 10),
        ([[SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_2, 10]], True, 10),
        ([["not_a_cycle", 10]], True, None),
    ],
    ids=["index2_cycles", "index2_cycles_nullable", "not_a_cycle_nullable"],
)
def test_get_cycle(
    sample_sheet_validator: SampleSheetValidator,
    sample_sheet_content: list[list[str]],
    nullable: bool,
    expected: int,
):
    """Test that a cycle value is fetched when the content is correct."""
    # GIVEN a sample sheet validator with a modified sample sheet content
    sample_sheet_validator.set_sample_sheet_content(sample_sheet_content)

    # WHEN fetching the cycle value
    result = sample_sheet_validator._get_cycle(
        cycle_name=SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_2, nullable=nullable
    )

    # THEN the correct value is returned
    assert result == expected


def test_get_cycle_missing_cycle(sample_sheet_validator: SampleSheetValidator, caplog):
    """Test that fetching a missing cycle value when nullable is False fails."""
    # GIVEN a sample sheet validator with a modified sample sheet content
    sample_sheet_validator.set_sample_sheet_content([["not_a_cycle", 10]])

    # WHEN fetching the cycle value
    with pytest.raises(SampleSheetError):
        # THEN a SampleSheetError is raised
        sample_sheet_validator._get_cycle(
            cycle_name=SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_2, nullable=False
        )
    assert (
        f"No {SampleSheetBCLConvertSections.Reads.INDEX_CYCLES_2} found in sample sheet"
        in caplog.text
    )


@pytest.mark.parametrize(
    "sample_sheet_content, expected_cycles",
    [
        ("hiseq_x_single_index_sample_sheet_content", [151, 151, 8, None]),
        ("hiseq_x_dual_index_sample_sheet_content", [151, 151, 8, 8]),
        ("hiseq_2500_dual_index_sample_sheet_content", [101, 101, 8, 8]),
        ("hiseq_2500_custom_index_sample_sheet_content", [101, 101, 17, 8]),
        ("novaseq_6000_pre_1_5_kits_sample_sheet_content", [151, 151, 10, 10]),
        ("novaseq_6000_post_1_5_kits_sample_sheet_content", [151, 151, 10, 10]),
        ("novaseq_x_sample_sheet_content", [151, 151, 10, 10]),
    ],
    ids=[
        "HiSeqXSingleIndex",
        "HiSeqXDualIndex",
        "HiSeq2500DualIndex",
        "HiSeq2500CustomIndex",
        "NovaSeq6000Pre1.5Kits",
        "NovaSeq6000Post1.5Kits",
        "NovaSeqX",
    ],
)
def test_set_cycles(
    sample_sheet_validator: SampleSheetValidator,
    sample_sheet_content: str,
    expected_cycles: list[int | None],
    request: FixtureRequest,
    caplog,
):
    """Test that the correct values for the cycles are set."""
    # GIVEN a valid sample sheet content
    content: list[list[str]] = request.getfixturevalue(sample_sheet_content)
    sample_sheet_validator.set_sample_sheet_content(content)

    # GIVEN a sample sheet validator with no cycles set
    assert sample_sheet_validator.read1_cycles is None
    assert sample_sheet_validator.read2_cycles is None
    assert sample_sheet_validator.index1_cycles is None
    assert sample_sheet_validator.index2_cycles is None

    # WHEN setting the cycles
    sample_sheet_validator._set_cycles()

    # THEN the correct values set as expected
    assert sample_sheet_validator.read1_cycles == expected_cycles[0]
    assert sample_sheet_validator.read2_cycles == expected_cycles[1]
    assert sample_sheet_validator.index1_cycles == expected_cycles[2]
    assert sample_sheet_validator.index2_cycles == expected_cycles[3]


@pytest.mark.parametrize(
    "sample_sheet_content",
    [
        "novaseq_6000_sample_sheet_with_reversed_cycles_content",
        "novaseq_x_sample_sheet_with_forward_cycles_content",
    ],
)
def test_validate_override_cycles_incorrect_cycles(
    sample_sheet_validator: SampleSheetValidator, sample_sheet_content: str, request: FixtureRequest
):
    """Test that a sample sheets with incorrect override cycles raise an error."""
    # GIVEN a sample sheet with incorrect override cycles
    sample_sheet_validator.set_sample_sheet_content(request.getfixturevalue(sample_sheet_content))

    # WHEN validating the override cycles
    with pytest.raises(SampleSheetError):
        # THEN a SampleSheetError is raised
        sample_sheet_validator._validate_override_cycles()


@pytest.mark.parametrize(
    "sample_sheet_content",
    [
        "hiseq_x_single_index_sample_sheet_content",
        "hiseq_x_single_index_bcl2fastq_sample_sheet_content",
        "hiseq_x_dual_index_sample_sheet_content",
        "hiseq_x_dual_index_bcl2fastq_sample_sheet_content",
        "hiseq_2500_dual_index_sample_sheet_content",
        "hiseq_2500_dual_index_bcl2fastq_sample_sheet_content",
        "hiseq_2500_custom_index_sample_sheet_content",
        "hiseq_2500_custom_index_bcl2fastq_sample_sheet_content",
        "novaseq_6000_pre_1_5_kits_sample_sheet_content",
        "novaseq_6000_post_1_5_kits_sample_sheet_content",
        "novaseq_x_sample_sheet_content",
    ],
    ids=[
        "HiSeqXSingleIndex",
        "HiSeqXSingleIndexBcl2Fastq",
        "HiSeqXDualIndex",
        "HiSeqXDualIndexBcl2Fastq",
        "HiSeq2500DualIndex",
        "HiSeq2500DualIndexBcl2Fastq",
        "HiSeq2500CustomIndex",
        "HiSeq2500CustomIndexBcl2Fastq",
        "NovaSeq6000Pre1.5Kits",
        "NovaSeq6000Post1.5Kits",
        "NovaSeqX",
    ],
)
def test_validate_sample_sheet_from_content(
    sample_sheet_validator: SampleSheetValidator, sample_sheet_content: str, request: FixtureRequest
):
    """Test that a correct sample sheet passes validation."""
    # GIVEN sample sheet validator and a correct sample sheet content
    content: list[list[str]] = request.getfixturevalue(sample_sheet_content)

    # WHEN validating the sample sheet
    sample_sheet_validator.validate_sample_sheet_from_content(content)

    # THEN no error is raised
